//************************************************************************************************************************** 
//* Blue HttpMail Extension
//* Copyright (c) 2003-2004 Josh Harler
//* 
//* Blue - General Purpose C++ Library
//* Copyright (c) 2002-2004 Josh Harler
//* 
//* This software is provided 'as-is', without any express or implied warranty. In no event
//* will the authors be held liable for any damages arising from the use of this software.
//* 
//* Permission is granted to anyone to use this software for any purpose, including commercial
//* applications, and to alter it and redistribute it freely, subject to the following restrictions:
//* 
//* 	1. The origin of this software must not be misrepresented; you must not claim that you
//* 	wrote the original software. If you use this software in a product, an acknowledgment in the
//* 	product documentation would be appreciated but is not required.
//* 
//* 	2. Altered source versions must be plainly marked as such, and must not be misrepresented as
//* 	being the original software.
//* 
//* 	3. This notice may not be removed or altered from any source distribution.
//*
//*
//* file   Blue/Extension/HttpMail/Folder.cpp
//**

// Private Headers =========================================================================================================

// matching header
#include "Folder.h"

// Extension headers
#include "Blue/Extension/HttpMail/Mailbox.h"

// Blue library headers
#include "Blue/Data/BufferDevice.h"

// Xml Extension headers
#include "Blue/Extension/Xml/DomParser.h"
#include "Blue/Extension/Xml/XPathExpression.h"


// Private Defines/Enums/Typedefs/Etc ======================================================================================

using namespace blue;
using namespace blue::common;
using namespace blue::data;
using namespace blue::ext;
using namespace blue::ext::xml;
using namespace blue::ext::http;


// Private Classes/Structs =================================================================================================

// Private Global Variables ================================================================================================

namespace {

	const String g_reqFolderList = 
		$("<?xml version=\"1.0\"?>\n"
		  "<D:propfind xmlns:D=\"DAV:\" xmlns:hm=\"urn:schemas:httpmail:\">\n"
		  "	<D:prop>\n"
		  "		<D:isfolder/>\n"
		  "		<D:displayname/>\n"
		  "		<hm:special/>\n"
		  "		<D:hassubs/>\n"
		  "		<D:nosubs/>\n"
		  "		<hm:unreadcount/>\n"
		  "		<D:visiblecount/>\n"
		  "		<hm:special/>\n"
		  "	</D:prop>\n"
		  "</D:propfind>\n");

	const String g_reqFolderInfo =
		$("<?xml version=\"1.0\"?>\n"
		  "<D:propfind xmlns:D=\"DAV:\" xmlns:hm=\"urn:schemas:httpmail:\" xmlns:m=\"urn:schemas:mailheader:\">\n"
		  "	<D:prop>\n"
		  "		<D:id/>\n"
		  "		<hm:read/>\n"
		  "		<hm:len/>\n"
		  "		<m:from/>\n"
		  "		<h:fromaddr/>\n"
		  "		<m:date/>\n"
		  "		<m:subject/>\n"
		  "	</D:prop>\n"
		  "</D:propfind>\n");

	const String g_setProperty =
		$("<?xml version=\"1.0\"?>\n"
		  "<D:propertyupdate xmlns:D=\"DAV:\" xmlns:hm=\"urn:schemas:httpmail:\">\n"
		  "	<D:set>\n"
		  "		<D:prop>\n"
		  "			<{1}>{2}</{1}>\n"
		  "		</D:prop>\n"
		  "	</D:set>\n"
		  "</D:propertyupdate>\n");
		
}


// External Global Variables ===============================================================================================

// Private Functions =======================================================================================================

// Functions ===============================================================================================================

namespace blue {
namespace ext {
namespace httpmail {

	// ---------------------------------------------------------------------------------------------------------------------

	Folder::Folder( Mailbox& mailbox, Folder* parent, String url )
		:m_mailbox(mailbox), m_parent(parent), m_url(url)
	{
		m_msgCount       = 0;
		m_msgCountUnread = 0;
	}

	// ---------------------------------------------------------------------------------------------------------------------

	Folder::~Folder()
	{
		for( int iF = 0; iF < m_folders.getSize(); ++iF ) {
			delete m_folders[iF];
		}

		for( int iM = 0; iM < m_messages.getSize(); ++iM ) {
			delete m_messages[iM];
		}
	}

	// ---------------------------------------------------------------------------------------------------------------------

	String Folder::getUrl() const
	{
		return (m_url);
	}

	// ---------------------------------------------------------------------------------------------------------------------

	String Folder::getName() const
	{
		return (m_name);
	}

	// ---------------------------------------------------------------------------------------------------------------------

	String Folder::getDisplayName() const
	{
		// this needs removed:
		if( m_display == String::null && m_name != String::null ) {
			if( m_name.equalsIgnoreCase("msnpromo") ) {
				return $("MSN Promotional");
			}
			if( m_name.equalsIgnoreCase("inbox") ) {
				return $("Inbox");
			}
			if( m_name.equalsIgnoreCase("sentitems") ) {
				return $("Sent Items");
			}
			if( m_name.equalsIgnoreCase("deleteditems") ) {
				return $("Deleted Items");
			}
			if( m_name.equalsIgnoreCase("bulkmail") ) {
				return $("Junk Mail");
			}
		}
		return (m_display);
	}

	// ---------------------------------------------------------------------------------------------------------------------

	int Folder::getMessageCount() const
	{
		if( m_messages.getSize() == 0 ) {
			return (m_msgCount);
		}
		else {
			return m_messages.getSize();
		}
	}

	// ---------------------------------------------------------------------------------------------------------------------

	int Folder::getMessageCountUnread() const
	{
		if( m_messages.getSize() == 0 ) {
			return (m_msgCountUnread);
		}
		else {
			int count = 0;
			for( int i = 0; i < m_messages.getSize(); ++i ) {
				if( m_messages[i]->isRead() == false ) {
					++count;
				}
			}
			return (count);
		}
	}

	// ---------------------------------------------------------------------------------------------------------------------

	int Folder::getFolderCount() const
	{
		return m_folders.getSize();
	}

	// ---------------------------------------------------------------------------------------------------------------------

	Folder* Folder::getFolder( int index )
	{
		if( index < 0 || index > m_folders.getSize()-1 ) {
			throw FolderException($("Unable to get folder: index out of bounds"));
		}
		return m_folders[index];
	}

	// ---------------------------------------------------------------------------------------------------------------------

	const Folder* Folder::getFolder( int index ) const
	{
		return ((Folder*)this)->getFolder(index);
	}

	// ---------------------------------------------------------------------------------------------------------------------

	Folder* Folder::getFolder( String folderName )
	{
		for( int iF = 0; iF < m_folders.getSize(); ++iF ) {
			if( m_folders[iF]->getName() == folderName ) {
				return (m_folders[iF]);
			}
		}

		return (0);
	}

	// ---------------------------------------------------------------------------------------------------------------------

	const Folder* Folder::getFolder( String folderName ) const
	{
		return ((Folder*)this)->getFolder(folderName);
	}

	// ---------------------------------------------------------------------------------------------------------------------

	Array<Folder*> Folder::getFolders()
	{
		return (m_folders);
	}

	// ---------------------------------------------------------------------------------------------------------------------

	const Array<Folder*> Folder::getFolders() const
	{
		return (m_folders);
	}

	// ---------------------------------------------------------------------------------------------------------------------

	Message* Folder::getMessage( int index )
	{
		if( index < 0 || index > m_messages.getSize()-1 ) {
			throw FolderException($("Unable to get message: index out of bounds"));
		}
		return (m_messages[index]);
	}
	
	// ---------------------------------------------------------------------------------------------------------------------
	
	const Message* Folder::getMessage( int index ) const
	{
		return ((Folder*)this)->getMessage(index);
	}
	
	// ---------------------------------------------------------------------------------------------------------------------
	
	Array<Message*> Folder::getMessages()
	{
		return (m_messages);
	}
	
	// ---------------------------------------------------------------------------------------------------------------------
	
	const Array<Message*> Folder::getMessages() const
	{
		return (m_messages);
	}
	
	// ---------------------------------------------------------------------------------------------------------------------

	Folder* Folder::getParent()
	{
		return (m_parent);
	}
	
	// ---------------------------------------------------------------------------------------------------------------------
	
	const Folder* Folder::getParent() const
	{
		return (m_parent);
	}
	
	// ---------------------------------------------------------------------------------------------------------------------

	void Folder::setUrl( String url )
	{
		m_url = url;
	}

	// ---------------------------------------------------------------------------------------------------------------------

	void Folder::setName( String name )
	{
		m_name = name;
	}

	// ---------------------------------------------------------------------------------------------------------------------

	void Folder::setDisplay( String display )
	{
		m_display = display;
	}

	// ---------------------------------------------------------------------------------------------------------------------

	void Folder::setMessageCount( int messageCnt )
	{
		m_msgCount = messageCnt;
	}

	// ---------------------------------------------------------------------------------------------------------------------

	void Folder::setMessageCountUnread( int messageCntUnread )
	{
		m_msgCountUnread = messageCntUnread;
	}

	// ---------------------------------------------------------------------------------------------------------------------

	int Folder::createFolder( String folderName )
	{
		String url = m_url;
		if( url.beginsWith("root:") ) {
			url = url.stripFromLeft(5);
		}

		if( !url.endsWith("/") ) {
			url += "/";
		}

		url += folderName;

		String errMsg;
		m_mailbox.processWebDavRequest(Url(url), $("MKCOL"), String::null, errMsg);
		if( errMsg != String::null ) {
			throw FolderException("Unable to create folder " + folderName + ": " + errMsg);
		}

		Folder* folder = new Folder(m_mailbox, this, url);
		folder->m_name    = folderName;
		folder->m_display = folderName;

		for( int i = 0; i < m_folders.getSize(); ++i ) {
			if( m_folders[i]->m_display.compareIgnoreCase(folderName) < 0 ) {
				m_folders.insert(folder, i);
				return (i);
			}
		}

		m_folders.append(folder);
		return (m_folders.getSize() - 1);
	}

	// ---------------------------------------------------------------------------------------------------------------------

	void Folder::removeFolder( int index )
	{
		if( Folder* folder = getFolder(index) ) {
			String url = folder->getUrl();

			String errMsg;
			String result = m_mailbox.processWebDavRequest(Url(url), $("DELETE"), String::null, errMsg);
			if( result == String::null ) {
				throw FolderException("Unable to remove folder " + folder->getName() + ": " + errMsg);
			}

			m_folders.remove(index);
		}
	}

	// ---------------------------------------------------------------------------------------------------------------------

	void Folder::removeFolder( String folderName )
	{
		for( int iF = 0; iF < m_folders.getSize(); ++iF ) {
			if( m_folders[iF]->getName() == folderName ) {
				removeFolder(iF);
				break;
			}
		}
	}

	// ---------------------------------------------------------------------------------------------------------------------

	void Folder::moveMessage( int index, Folder* to )
	{
		if( &to->m_mailbox != &m_mailbox ) {
			throw FolderException($("Messages can only be moved to folders in the same mailbox"));
		}

		Message* message = getMessage(index);
		bool read = message->isRead();

		String url = to->m_url + "/" + message->getID();

		Headers headers;
		headers.addField("Destination", url);

		String errMsg;
		m_mailbox.processWebDavRequest(Url(message->getUrl()), $("MOVE"), String::null, headers, errMsg);
		if( errMsg != String::null ) {
			throw FolderException("Unable to move message to folder: " + errMsg);
		}

		m_messages.remove(index);
		delete message;

		m_msgCount -= 1;
		if( read ) {
			m_msgCountUnread -= 1;
		}
	}

	// ---------------------------------------------------------------------------------------------------------------------

	void Folder::moveMessage( Message* message, Folder* to )
	{
		for( int i = 0; i < m_messages.getSize(); ++i ) {
			if( m_messages[i] == message ) {
				moveMessage(i, to);
				break;
			}
		}
	}

	// ---------------------------------------------------------------------------------------------------------------------

	void Folder::copyMessage( int index, Folder* to )
	{
		if( &to->m_mailbox != &m_mailbox ) {
			throw FolderException($("Messages can only be copied to folders in the same mailbox"));
		}

		Message* message = getMessage(index);

		String url = to->m_url + "/" + message->getID();

		Headers headers;
		headers.addField("Destination", url);

		String errMsg;
		m_mailbox.processWebDavRequest(Url(message->getUrl()), $("COPY"), String::null, headers, errMsg);
		if( errMsg != String::null ) {
			throw FolderException("Unable to copy message to folder: " + errMsg);
		}
	}

	// ---------------------------------------------------------------------------------------------------------------------

	void Folder::copyMessage( Message* message, Folder* to )
	{
		for( int i = 0; i < m_messages.getSize(); ++i ) {
			if( m_messages[i] == message ) {
				copyMessage(i, to);
				break;
			}
		}
	}

	// ---------------------------------------------------------------------------------------------------------------------

	void Folder::removeMessage( int index )
	{
		Folder* folder = m_mailbox.getRootFolder()->getFolder($("deleteditems"));
		if( folder == 0 ) {
			throw FolderException($("Unable to locate 'Deleted Items' folder"));
		}
		if( folder != this ) { // if not already deleted
			moveMessage(index, folder);
		}
	}

	// ---------------------------------------------------------------------------------------------------------------------

	void Folder::removeMessage( Message* message )
	{
		Folder* folder = m_mailbox.getRootFolder()->getFolder($("deleteditems"));
		if( folder == 0 ) {
			throw FolderException($("Unable to locate 'Deleted Items' folder"));
		}
		if( folder != this ) { // if not already deleted
			for( int i = 0; i < m_messages.getSize(); ++i ) {
				if( m_messages[i] == message ) {
					moveMessage(i, folder);
					break;
				}
			}
		}
	}

	// ---------------------------------------------------------------------------------------------------------------------

	void Folder::setMessageRead( int index, bool read )
	{
		Message* message = getMessage(index);

		String propXml = g_setProperty.arg("hm:read").arg((read ? "1" : "0"));

		String errMsg;
		m_mailbox.processWebDavRequest(Url(message->getUrl()), $("PROPPATCH"), propXml, errMsg);
		if( errMsg != String::null ) {
			throw FolderException("Unable to mark message as read: " + errMsg);
		}

		message->setRead(read);
	}

	// ---------------------------------------------------------------------------------------------------------------------

	void Folder::setMessageRead( Message* message, bool read )
	{
		for( int i = 0; i < m_messages.getSize(); ++i ) {
			if( m_messages[i] == message ) {
				setMessageRead(i, read);
				break;
			}
		}
	}

	// ---------------------------------------------------------------------------------------------------------------------

	void Folder::refresh()
	{
		bool root = m_url.beginsWith("root:");

		Url url;
		String content;

		if( root ) {
			url.setUrl(m_url.stripFromLeft(5));
			content = g_reqFolderList;
		}
		else {
			url.setUrl(m_url);
			content = g_reqFolderInfo;
		}

		String errMsg;
		String result = m_mailbox.processWebDavRequest(url, $("PROPFIND"), content, errMsg);

		if( result == String::null ) {
			throw FolderException("Unable to get folder information: '" + errMsg + "'");
		}

		BufferDevice device(result);
		InputStream parserStream(&device, true);
		DomParser parser(&parserStream);
		parser.parse();

		Array<DomNode*> folderNodes = XPathExpression(parser.getDocumentNode()).findNodes($("//D:response"));
		Array<Folder*> newFolders;

		m_msgCount = 0;
		m_msgCountUnread = 0;

		for( int i = 0; i < folderNodes.getSize(); ++i ) {
			DomNode* node = folderNodes[i];
			XPathExpression xpath(node);
			String href = xpath.findValue($("/D:href"));

			if( xpath.findValue($("//D:isfolder")) == "1" ) {
				Folder* folder = 0;
				for( int iF = 0; iF < m_folders.getSize(); ++iF ) {
					if( m_folders[iF]->m_url == href ) {
						folder = m_folders[iF];
						break;
					}
				}
				if( folder == 0 ) {
					folder = new Folder(m_mailbox, this, href);
					m_folders.append(folder);
				}

				folder->m_name    = xpath.findValue($("//* [name()='D:displayname' or name()='hm:special'][string-length(text()) > 0]"));
				folder->m_display = xpath.findValue($("//D:displayname"));
				
				String msgCountStr = xpath.findValue($("//D:visiblecount"));
				if( msgCountStr.isValidInt() ) {
					folder->m_msgCount = msgCountStr.getAsInt();
				}
				String msgCountUnreadStr = xpath.findValue($("//hm:unreadcount"));
				if( msgCountUnreadStr.isValidInt() ) {
					folder->m_msgCountUnread = msgCountUnreadStr.getAsInt();
				}
			}
			else {
				String messageID = xpath.findValue($("//D:id"));
				Message* message = 0;

				for( int iM = 0; iM < m_messages.getSize(); ++iM ) {
					if( m_messages[iM]->getID() == messageID ) {
						message = m_messages[iM];
						break;
					}
				}
				if( message == 0 ) {
					message = new Message(m_mailbox, href);
					m_messages.append(message);
					m_msgCount += 1;
				}

				String from     = xpath.findValue($("//m:from"));
				String fromAddr = xpath.findValue($("//h:fromaddr"));

				if( from == String::null && fromAddr == String::null && m_name.equalsIgnoreCase("sentitems")) {
					fromAddr = m_mailbox.getUser() + "@" + m_mailbox.getDomain();
				}
				
				bool read = xpath.findValue($("//hm:read")) == "1";
				if( read ) {
					m_msgCountUnread += 1;
				}
				message->setRead( read );
				message->setFrom( (from == "" || from == fromAddr ? fromAddr : "\"" + from + "\" <" + fromAddr + ">") );
				message->setSubject( xpath.findValue($("//m:subject")) );
				message->setContentLength( xpath.findValue($("//hm:len")).getAsInt() );

				String dateStr = xpath.findValue($("//m:date"));
				if( dateStr != String::null ) {
					// format: yyyy-mm-ddThh:mm:ss"
					util::DateTime date;
					date.setYear  (dateStr.subString( 0,4).getAsInt());
					date.setMonth (dateStr.subString( 5,2).getAsInt());
					date.setDay   (dateStr.subString( 8,2).getAsInt());
					date.setHour  (dateStr.subString(11,2).getAsInt());
					date.setMinute(dateStr.subString(14,2).getAsInt());
					date.setSecond(dateStr.subString(17,2).getAsInt());
					message->setDate(date);
				}
			}
		}
	}

}}}	// namespaces
